#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""
Program that sequentially launches phenix.ligand_pipeline on several sets of the same protein but with different ligands.
Specially adapted for ligand or fragment screening.
Argument: data directory

Subcommands:
* prep: preparation of the data = first step mandatory
Creation of a treatment directory: easYPipe, at the root of the datasets directory.
Creation of a subdirectory '0_processed_datasets' where all the datasets are copied, creation of a 'data' folder in each 
    dataset, copy of mtz process adding a mtz number.
If there are several mtz in a folder, searches for 'EDNA' treatment template and selects the right mtz file. 
If just one mtz in the folder, just copy it.
All informations on mtz files are written in 'mtz_to_treat_ALL.csv' in a 'prep' folder.

* reindex (optional): reindexation of mtz files to space group of a reference mtz
Argument: mtz of reference for reindexation 
Write a new 'mtz_to_treat_ALL.csv' with reindexed mtz files information.

* ligands: preparation for ligand search, generates pdb and cif of ligands in each dataset (in a 'ligand' folder),
        accordingly to the lig-csv input
Argument: ligands_for_datasets.csv file from 'prep' modified with ligands names and smiles      
Creates a smiles file in a ligand subdirectory of copied dataset and converts it with elBOW to pdb and cif.
 
* launch: 
Run phenix.ligand_pipeline on all the mtz according to options and information in 'mtz_to_treat_ALL.csv' file.
Argument: folder with fasta file and pdb file for replacement, and cif(s) if ligand(s) in the model
Modes (-m, --mode):
- fast (= default)
- full 
- full-lig
Launch 'full' if no ligand cif
- allsg
- allsg-lig
Launch 'allsg' if no ligand cif
Options:
- autoproc (--autoproc, -a): treat process from autoPROC, if none treat like '--best 1'
- best + nb (--best, -b) / default = '--best 1' : i.e. by default treat only the mtz with best completeness
- simulate (--simulate, -s): only simulates, generates (and replaces it if exists) a csv file according to the future launch options. 
Give the possibility to modify the csv file to choose not to launch certain treatments, 
before restarting without simulation mode.
- cclig (--cclig, -c) / default = 0.7 : Minimum CC to consider a ligand placement correct. 
Ligands with at least this CC will be incorporated into the current model for refinement.
- nblig (--nblig", -n) / default = 1 : Number of ligand copies to be searched (max 9 for the moment, cf bug Phenix 2021-01)."
- name (--name, -n): optional name for result folders, in case re-launching with different reference pdb 
of the same space group (because reference pdb space group is put in the results folders, but not cell parameters).

* pandda (--pandda, -p): copy results from easYPipe to a 'PANDDA' folder, as data ready to launch PanDDA


cf logs files in each folder ...
------------------------------------------------------------------------------------------------------------------------
phenix.ligand_pipeline options:
# common
nproc=Auto
preserve_chain_id=True : The default behavior is to group heteroatoms with the nearest macromolecule chain, whose ID is inherited. 
                        This parameter disables the change of chain ID, and preserves the original chain ID
refine.after_ligand.hydrogens=False : Hydrogen atoms won't be added prior to the final refinement step (else refinement significantly slower)
prune=False :disable Prune the model after refinement to remove residues and sidechains in poor density
keep_hetatms=True : to prevent Phaser from resetting the occupancies to zero => if True, HETATMs will not have their occupancies reset to zero.
refine.after_mr.update_waters=False => don't add/remove waters automatically (to better see ligand density).

# mode FAST
skip_xtriage=True
mr=False : rigid-body refinement will be used
quick_refine=True : which will shorten both refinement steps from 6 to 3 cycles, and disable weight optimization. Suitable for isomorphous structures.
build=False   (default=Auto => Run AutoBuild after initial refinement. By default, this will be done if R-free is greater than the max_r_free cutoff = 0.3)
skip_ligand=True
reference_structure='model.pdb' : If specified, phenix.find_alt_orig_sym_mate will be applied to map the solution to the reference structure 
                                    => plante avec Phaser si plusieurs monomères => enlevé pour allsg et full
                                             
# mode full 
mr=Auto : the program will try rigid-body refinement first, then run Phaser if the R-free is too high (>0.4)
build=Auto   (default=Auto => Run AutoBuild after initial refinement. By default, this will be done if R-free is greater than the max_r_free cutoff = 0.3)
autobuild.quick=True   (Run AutoBuild in quick mode. Inferior results, but a huge time-saver.)
quick_refine=True (Use a shorter refinement protocol: fewer cycles and no weight optimization.)
                             
# mode allsg
mr=True
quick_refine=False

# ligand                                              
ligand_copies=1 (except if args.nblig >1, and <=9 because bug else)
keep_input_restraints=True : if the input files include pre-calculated restraints for the target ligand, 
                            eLBOW will propagate these restraints instead of generating new ones.

"""


# ----------------------- Imports --------------------------------
import sys
from pathlib import Path
import time
import argparse

# my modules:
import config
import modules.subcommands
from modules.subcommands import *
from modules.utils import verbose
from modules.utils import proteins



# ----------------------- Functions-------------

def parse_arguments():
    """Parser"""
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description='''
    ---------------------------------------------------------------
    Program that sequentially launches phenix.ligand_pipeline
    on several sets of the same protein but with different ligands.
    Specially adapted for ligand or fragment screening.
    ---------------------------------------------------------------''')
    parser.add_argument("data", help="folder with datasets in subfolders")
    subparsers = parser.add_subparsers(title='subcommands', description="Valid subcommands", dest="subcommands")
    subparsers.required = True
    parser_prep = subparsers.add_parser('prep', help='prepare data: listing of mtz to treat with information')
    parser_reindex = subparsers.add_parser('reindex', help='try to reindex mtz according to the space group of reference file')
    parser_reindex.add_argument("sg_ref", help="space group of reference for reindexing (name or number)")
    parser_reindex.add_argument("-s", "--simulate", help="only simulate, generate a csv file listing for each process if mtz file will be "
                                                        "reindexed or not. Give the possibility to modify the csv file to choose not to launch "
                                                        "some reindexation, before launching again without simulation mode.",
                               action="store_true")
    parser_ligands = subparsers.add_parser('ligands',
                                           help='generate ligands files before launching with ligand search')
    parser_ligands.add_argument('ligands_csv',
                                help="ligands_for_datasets.csv file from 'prep' with ligands names and smiles completed")
    parser_launch = subparsers.add_parser('launch',
                                          help='launch all the "phenix.ligand-pipeline", after the preparation steps')
    parser_launch.add_argument("ref",
                               help="folder with fasta file and pdb file(s) for replacement, and cif(s) if ligand(s) in the model")
    parser_launch.add_argument("-m", "--mode", choices=['fast', 'full', 'allsg'],
                               default='fast',
                               help="running mode: fast, full, or allsg (default = fast)")
    parser_launch.add_argument("-l", "--lig", help="for ligand search and placement", action="store_true")
    parser_launch.add_argument("-n", "--nblig", type=int, default=1, metavar='NUMBER',
                               help="number of ligand copies to be "
                                    "searched (default = 1, max 9 for the moment).")
    parser_launch.add_argument("-c", "--cclig", type=float, default=0.7, metavar='NUMBER',
                               help="minimum CC to consider a "
                                    "ligand placement correct (default = 0.7). Ligands with at least this CC will be "
                                    "incorporated into the current model for refinement.")
    group_launch = parser_launch.add_mutually_exclusive_group()
    group_launch.add_argument("-b", "--best", type=int, default=1, nargs='?', const=1, metavar='NUMBER',
                              help="launch only for mtz with best completeness, "
                                   "NUMBER indicates how many mtz to treat (default 1), "
                                   "ex: -- best 2")
    group_launch.add_argument("-a", "--autoproc", help="launch only for mtz from autoPROC, or if none "
                                                       "launch for mtz with best completeness", action="store_true")
    group_launch.add_argument("-w", "--whole", help="launch for the whole mtz processes", action="store_true")
    parser_launch.add_argument("-s", "--simulate", help="only simulate, generate a csv file according to the future "
                                "launch options. Give the possibility to modify the csv file to choose not to launch "
                                                        "some treatments, before launching again without simulation mode.",
                               action="store_true")

    parser_launch.add_argument("-t", "--template", type=str, help="optional template name for log files and result folders, "
                                                                "in case re-launching with different reference pdb "
                                                              "of the same space group (else will be treated in existing "
                                                                  "folder and not launched again since it already exists).")
    parser_summary = subparsers.add_parser('summary',
                                          help="compile the results of all the 'launch' executions in a single csv file")
    parser_summary.add_argument("-o", "--only_success", help="generate also a result file with only successfull treatments", action="store_true")
    parser_auto = subparsers.add_parser('auto',help="run main easYPipe steps (prep, reindex, launch, summary) in automatic mode, "
                                               "but without ligand search")
    parser_auto.add_argument("ref", help="folder with fasta file and pdb file(s) for replacement, and cif(s) "
                                           "if ligand(s) in the model")
    parser_auto.add_argument("-m", "--mode", choices=['fast', 'full', 'allsg'],
                               default='fast',
                               help="running mode: fast, full, or allsg (default = fast)")
    group_auto = parser_auto.add_mutually_exclusive_group()
    group_auto.add_argument("-b", "--best", type=int, default=1, metavar='NUMBER',
                              help="launch only for mtz with best completeness, "
                                   "NUMBER indicates how many mtz to treat (default 1), "
                                   "ex: -- best 2")
    group_auto.add_argument("-a", "--autoproc", help="launch only for mtz from autoPROC, or if none "
                                                       "launch for mtz with best completeness", action="store_true")
    group_auto.add_argument("-w", "--whole", help="launch for the whole mtz processes", action="store_true")
    parser_auto.add_argument("-t", "--template", type=str,
                               help="optional template name for log files and result folders, "
                                    "in case re-launching with different reference pdb "
                                    "of the same space group (else will be treated in existing "
                                    "folder and not launched again since it already exists).")

    parser_pandda = subparsers.add_parser('pandda', help="copy results from easYPipe to a 'PANDDA' folder, "
                                                         "as data ready to launch PanDDA")
    return parser.parse_args()


#----------------------Main program------------------------

if __name__ == "__main__":

    start = time.process_time()
    start_perf = time.perf_counter()

    """Analyse of parser"""
    args = parse_arguments()

    """Check if data directory exists"""
    if Path(args.data).is_dir():
        data_dir = Path(args.data).resolve()    # chemin absolu
    else:
        sys.exit("Data directory not found")

    """Folder where we work"""
    work_dir = Path.cwd()


    """prog subcommand = prep or auto"""
    if args.subcommands == 'prep' or args.subcommands == 'auto':
        log = prep.preparation_run(data_dir, args)

    """prog subcommand = auto or launch => check ref folder and get ref space group"""
    if args.subcommands == 'launch' or args.subcommands == 'auto':

        """Check if reference folder and its content are OK"""
        list_pdb = []  # liste des pdb dans le dossier de réf
        list_fasta = []  # liste des fasta dans le dossier de réf
        list_refcif = []  # liste des cif dans le dossier de réf
        if Path(args.ref).is_dir():
            # vérifier qu'il y a un pdb, un fasta et éventuellement un cif
            files = [f for f in Path(args.ref).resolve().iterdir() if f.is_file()]
            list_pdb = [f for f in files if f.resolve().suffix.lower() == '.pdb']
            list_fasta = [f for f in files if f.resolve().suffix.lower() in ['.fas', '.fa', '.fasta']]
            list_refcif = [f for f in files if f.resolve().suffix.lower() == '.cif']
            if not list_pdb:
                sys.exit("No reference pdb file found")
            if not list_fasta:
                sys.exit("No reference fasta file found")
            if len(list_fasta) > 1:
                sys.exit("More than one fasta file in the reference folder.")
        else:
            sys.exit("Reference folder not found")

        """Get reference space group in reference pdb files"""
        sg_ref = ""
        if args.mode != 'allsg':
            list_sg = []
            for pdb in list_pdb:
                print("Searching space group in " + pdb.name)
                sg = proteins.sg_from_pdb(pdb)
                list_sg.append(sg)
                print("Space group: " + sg)
            if list_sg.count("") == len(list_pdb):  # aucun sg trouvé
                sys.exit("No space group found in reference pdb file" + ("s." if len(list_pdb) > 1 else "."))
            if list_sg.count("") != 0 and list_sg.count("") != len(
                    list_pdb):  # i.e. certain sg n'ont pas été trouvés, certains oui
                # on écarte les pdb correspondant car pourraient poser problème
                for i, sg in enumerate(list_sg):
                    if sg == "":
                        del list_pdb[i]
                        del list_sg[i]
            if len(list_pdb) == len(list_sg) and list_sg.count(list_sg[0]) == len(
                    list_sg):  # les sg de chaque pdb ont été trouvés
                # et sont idem
                sg_ref = list_sg[0]
                print("Reference space group: " + sg_ref)
            else:
                print("Reference space groups found in pdb files: " + ','.join(list_sg))
                sys.exit("Reference pdb files have not the same space group.")


    """prog subcommand = reindex or auto"""
    if args.subcommands == 'reindex' or args.subcommands == 'auto':

        if args.subcommands == 'reindex':   # si 'auto', le sg_ref est récupéré dans les pdb de ref, cf plus haut
            """Check if treatment directory and csv file with info on mtz exist"""
            phenix_dir = work_dir.joinpath('easYPipe')
            prep_dir = phenix_dir.joinpath('1a_prep')
            if not prep_dir.exists():
                sys.exit("Please run '" + Path(sys.argv[0]).name + " prep' before")
            mtz_treat_csv = prep_dir.joinpath('mtz_to_treat_ALL.csv')
            if not mtz_treat_csv.exists():
                sys.exit("Please run '" + Path(sys.argv[0]).name + " prep' before.\n Please do not change anything "
                                                                   "to 'mtz_to_treat_ALL.csv' file created during preparation.")

            """Check if reference space group exists"""
            sg_ref = proteins.clean_sg_name(args.sg_ref)
            if not sg_ref:
                # test si numéro à la place du nom
                sg_ref = proteins.sg_nb_to_symbol.get(int(args.sg_ref.strip()), 0)  # return 0 if the specified key does not exist
                if sg_ref == 0:
                    sys.exit("Reference space group for reindexing not correct: " + str(args.sg_ref))

        elif args.subcommands == 'auto': # subcommands pas dans auto
            args.simulate = False

        log = reindex.reindex_run(data_dir, args, sg_ref)



    """prog subcommand = ligands"""
    if args.subcommands == 'ligands':

        """Check if treatment directory exists"""
        phenix_dir = work_dir.joinpath('easYPipe')
        if not phenix_dir.exists():
            sys.exit("Please run '" + Path(sys.argv[0]).name + " prep' before")

        """Check if ligands csv file exists"""
        if Path(args.ligands_csv).is_file():
            lig_csv_file = Path(args.ligands_csv).resolve()   # pour avoir un path absolu!
        else:
            sys.exit("csv file for ligands not found")

        log = ligands.ligands_run(data_dir, lig_csv_file)



    """prog subcommand = launch (following ..., see above)"""
    if args.subcommands == 'launch' or args.subcommands == 'auto':

        if args.subcommands == 'launch':        # pas besoin de le vérifier si 'auto'
            """Check if treatment directory exists"""
            phenix_dir = work_dir.joinpath('easYPipe')
            if not phenix_dir.exists():
                sys.exit("Please run '" + Path(sys.argv[0]).name + " prep' before")

            """Check if csv file exists"""
            prep_dir = phenix_dir.joinpath('1a_prep')
            mtz_treat_csv = prep_dir.joinpath('mtz_to_treat_ALL.csv')
            if mtz_treat_csv not in [f for f in prep_dir.iterdir() if f.is_file()]:
                sys.exit("File 'mtz_to_treat_ALL.csv' not found.\nPlease run '" + Path(sys.argv[0]).name + " prep' before")

            """Check if cc ligand cutoff is correct"""
            if not (args.cclig > 0 and args.cclig <= 1):
                sys.exit("cclig must be between 0 and 1")

            """Check if number of ligand copies to search is <=9 (bug)"""
            if not (args.nblig <= 9):
                sys.exit("nblig must be between 1 and 9 (bug if more for the moment)")
        elif args.subcommands == 'auto': # subcommands pas dans auto
            args.lig = False
            args.nblig = False
            args.cclig = False
            args.simulate = False

        log = launch.launch_run(data_dir, args, sg_ref)

    """prog subcommand = summary or auto"""
    if args.subcommands == 'summary' or args.subcommands == 'auto':


        """Check if treatment directory exists"""
        phenix_dir = work_dir.joinpath('easYPipe')
        if not phenix_dir.exists():   # pas besoin si 'auto':
            sys.exit("No 'easYPipe' folder, please run '" + Path(sys.argv[0]).name + " prep, then launch' before")

        """Check if easypipe result directory exists"""
        # parcourir et rechercher les dossiers commençant par 'RESULTS_'
        res_dirs = [f.name for f in phenix_dir.iterdir() if f.is_dir() and f.name.startswith('RESULTS')]
        if not res_dirs:
            sys.exit(
                "No results folder in 'easYPipe' folder, please run '" + Path(sys.argv[0]).name + " launch' before")

        log = summary.summary_run(data_dir, res_dirs, args)

    """prog subcommand = pandda"""
    if args.subcommands == 'pandda':

        """Check if treatment directory exists"""
        phenix_dir = work_dir.joinpath('easYPipe')
        if not phenix_dir.exists():
            sys.exit("No 'easYPipe' folder, please run '" + Path(sys.argv[0]).name + " prep, then launch' before")
        """Check if easypipe result directory exists"""
        # parcourir et rechercher les dossiers commençant par 'RESULTS_'
        res_dirs = [f.name for f in phenix_dir.iterdir() if f.is_dir() and f.name.startswith('RESULTS')]
        if not res_dirs:
            sys.exit("No results folder in 'easYPipe' folder, please run '" + Path(sys.argv[0]).name + " launch' before")

        log = pandda.data_for_pandda(data_dir, res_dirs)


    """Written at the end of each log file"""
    # programmes utilisés
    verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs, "Software used in easYPipe:"), log)
    verbose.print_and_write(config.PHENIX_PATH, log)
    verbose.print_and_write("\tphenix.xtriage", log)
    verbose.print_and_write("\tphenix.elbow", log)
    verbose.print_and_write("\tphenix.ligand_pipeline", log)
    verbose.print_and_write(config.BABEL_PATH + "babel", log)


    # performance
    verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs, "Program performance"), log)
    end = time.process_time()
    end_perf = time.perf_counter()
    totalCPU = end - start
    total_perf = end_perf - start_perf
    #verbose.print_and_write("Duration / CPU (hours:minutes:seconds) = " + str(datetime.timedelta(seconds=totalCPU)), log)
    #verbose.print_and_write("Duration / real time (hours:minutes:seconds) = " + str(datetime.timedelta(seconds=total_perf)) + "\n", log)
    #verbose.print_and_write("Duration / CPU = " + "{0} hours : {1} minutes : {2} seconds"
    #      .format(*verbose.seconds_to_hhmmss(int(totalCPU))) + ("(" + str(totalCPU) + " seconds)" if totalCPU < 1 else ""), log)
    verbose.print_and_write("Duration = " + "{0} hours : {1} minutes : {2} seconds"
          .format(*verbose.seconds_to_hhmmss(int(total_perf))) + ("(" + str(total_perf) + " seconds)\n" if total_perf < 1 else "\n"), log)






